#include    <stdio.h>
#include	<stdlib.h>
#include	<conio.h>
#include	<dos.h>

#include	"stdtype.h"
#include	"config.h"
#include	"1938.h"
#include	"delay.h"
#include    "dsp.h"
#include    "mixer.h"
#include	"screenio.h"
#include	"display.h"
#include	"ic8237.h"


/*****************************************************************************/
/* bGetDSPStatus - Get Status from DSP                                       */
/*                                                                           */
/*****************************************************************************/
BYTE    bGetDSPStatus( void )
	{
	return ( ( BYTE ) inp( _ES_DSP_RW_STATUS ) );
	}

/*****************************************************************************/
/* bGetDSPRdStatus - Get DSP Read Status                                     */
/*                                                                           */
/*****************************************************************************/
BYTE    bGetDSPRdStatus( void )
	{
	return ( ( BYTE ) ( ( ( bGetDSPStatus(  ) & _ES_DSP_RD_MSK ) == _ES_DSP_RD_RDY ) ? 1 : 0 ) );
	}

/*****************************************************************************/
/* bGetDSPWrStatus - Get DSP Write Status                                    */
/*                                                                           */
/*****************************************************************************/
BYTE    bGetDSPWrStatus( void )
	{
	return ( ( BYTE ) ( ( ( bGetDSPStatus(  ) & _ES_DSP_WR_MSK ) == _ES_DSP_WR_RDY ) ? 1 : 0 ) );
	}

/*****************************************************************************/
/* iWrDSP - Write Data to DSP                                                */
/*                                                                           */
/*****************************************************************************/
int     iWrDSP( BYTE _bDSPData )
	{
    int     iRetVal = _ES_DSP_ERROR;

    BYTE	bTimeoutFlag = FALSE;
    
    vSetEventWait( _1mS * 5, &bTimeoutFlag );
    
	while ( bTimeoutFlag == FALSE )
		{
		if ( bGetDSPWrStatus() == TRUE )
			{
			outp( _ES_DSP_DATA_OUT, _bDSPData );
			iRetVal = _ES_DSP_OK;
			break;
			}
		}	
			
    vClearEventWait();

    return ( iRetVal );
    }

/*****************************************************************************/
/* iRdDSP - Read Data from DSP                                               */
/*                                                                           */
/*****************************************************************************/
int     iRdDSP( void )
	{
    int     iRetVal = _ES_DSP_ERROR;

	BYTE	bTimeoutFlag = FALSE;

    vSetEventWait( _1mS * 5, &bTimeoutFlag );

	while ( bTimeoutFlag == FALSE )
		{
		if ( bGetDSPRdStatus() == TRUE )
			{
			iRetVal = inp( _ES_DSP_DATA_IN );
			break;
			}
		}
		
    vClearEventWait();

	return ( iRetVal );
    }


/*****************************************************************************/
/* iRdXReg - Read DSP Extended Register                                      */
/*                                                                           */
/*****************************************************************************/
int     iRdXReg( BYTE _bXReg )
	{
	int	iRetVal = _ES_DSP_ERROR;
	
	if ( ( iRetVal = iWrDSP( 0xC0 ) ) != _ES_DSP_ERROR )
		if( ( iRetVal = iWrDSP( _bXReg ) ) != _ES_DSP_ERROR )
			iRetVal = iRdDSP();

	return ( iRetVal );
	}

/*****************************************************************************/
/* iWrXReg - Write DSP Extended Register                                     */
/*                                                                           */
/*****************************************************************************/
int     iWrXReg( BYTE _bXReg, BYTE _bXData )
	{
	int     iRetVal = _ES_DSP_ERROR;

	if( ( iRetVal = iWrDSP( _bXReg ) ) != _ES_DSP_ERROR )
		iRetVal = iWrDSP( _bXData );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iSetXRegBit - Set Bit in DSP Extended Register                            */
/*                                                                           */
/*****************************************************************************/
int     iSetXRegBit( BYTE _bXReg, BYTE _bBit )
	{
	int 	iRetVal = _ES_DSP_ERROR;

	int		iTmpVar;

	if( ( iTmpVar = ( BYTE ) iRdXReg( _bXReg ) ) != _ES_DSP_ERROR )
		{
		iRetVal = iWrXReg( _bXReg, BSETBIT( iTmpVar, _bBit ) );
		}

	return ( iRetVal );
	}

/*****************************************************************************/
/* iMskXRegBit - Clear Bit in DSP Extended Register                          */
/*                                                                           */
/*****************************************************************************/
int     iMskXRegBit( BYTE _bXReg, BYTE _bBit )
	{
	int 	iRetVal = _ES_DSP_ERROR;

	int		iTmpVar;

	if ( ( iTmpVar = ( BYTE ) iRdXReg( _bXReg ) ) != _ES_DSP_ERROR )
		{
		iRetVal = iWrXReg( _bXReg, BMSKBIT( iTmpVar, _bBit ) );
		}

	return ( iRetVal );
	}

/*****************************************************************************/
/* iTstXRegBit - Test Bit in DSP Extended Register                           */
/*                                                                           */
/*****************************************************************************/
int     iTstXRegBit( BYTE _bXReg, BYTE _bBit )
	{
	int	iRetVal = _ES_DSP_ERROR;

	int	iTmpVar;

	if ( ( iTmpVar = iRdXReg( _bXReg ) ) != _ES_DSP_ERROR )
		iRetVal = BTSTBIT( iTmpVar, _bBit );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iResetDSP - Reset DSP and Return DSP Status                               */
/*                                                                           */
/*****************************************************************************/
int     iResetDSP( void )
	{
	vDisplayDebugMessage( "iResetDSP\n" );

	outp( _ES_DSP_RESET, _ES_DSP_RESET_ON );
    vDelay( _1mS  );

	outp( _ES_DSP_RESET, _ES_DSP_RESET_OFF );
    vDelay( _1mS  );

	return ( iRdDSP(  ) );
	}

/*****************************************************************************/
/* iResetFIFO - Reset DSP FIFO & Return DSP Status                           */
/*                                                                           */
/*****************************************************************************/
int     iResetFIFO( void )
	{
	vDisplayDebugMessage( "iResetFIFO\n" );

	outp( _ES_DSP_RESET, _ES_DSP_RESET_FIFO_ON );
    vDelay( _1mS  );

    outp( _ES_DSP_RESET, _ES_DSP_RESET_FIFO_OFF );
    vDelay( _1mS  );

	return ( iRdDSP(  ) );
	}


/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
BYTE    bRdFIFO( void )
	{
	return ( ( BYTE ) inp( _ES_DSP_FIFO_DATA ) );
	}

/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
BYTE    bWrFIFO( BYTE _bData )
	{
	return ( ( BYTE ) outp( _ES_DSP_FIFO_DATA, _bData ) );
	}

/*****************************************************************************/
/* iEnableExtendedMode - Enable DSP Extended Registers & Functions           */
/*                                                                           */
/*****************************************************************************/
int     iEnableExtendedMode( void )
	{
	int     iRetVal = _ES_DSP_ERROR;

	vDisplayDebugMessage( "iEnableExtendedMode\n" );

	iRetVal = iWrDSP( _ES_DSP_XMODE_ON );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iDisableExtendedMode - Disable DSP Extended Registers & Functions         */
/*                                                                           */
/*****************************************************************************/
int     iDisableExtendedMode( void )
	{
	int     iRetVal = _ES_DSP_ERROR;

	vDisplayDebugMessage( "iDisableExtendedMode\n" );

	iRetVal = iWrDSP( _ES_DSP_XMODE_OFF );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iEnableRecordMonitor - Enable Audio Recording Monitor                     */
/*                                                                           */
/*****************************************************************************/
int     iEnableRecordMonitor( void )
	{
	int     iRetVal = _ES_DSP_ERROR;

	vDisplayDebugMessage( "iEnableRecordMonitor\n" );

	iRetVal = iSetXRegBit( 0xA8, 3 );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iDisableRecordMonitor - Disable Audio Recording Monitor                   */
/*                                                                           */
/*****************************************************************************/
int     iDisableRecordMonitor( void )
	{
	int     iRetVal = _ES_DSP_ERROR;

	vDisplayDebugMessage( "iDisableRecordMonitor\n" );

	iRetVal = iMskXRegBit( 0xA8, 3 );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iGetDACStatus - Return DAC Speaker Status (0=Off,1=On)                    */
/*                                                                           */
/*****************************************************************************/
int     iGetDACStatus( void )
	{
	int		iRetVal = _ES_DSP_ERROR;
	
	if( ( iRetVal = iWrDSP( 0xD8 ) ) != _ES_DSP_ERROR )
		{
		iRetVal = iRdDSP();
		}

	return( iRetVal );
	}

/*****************************************************************************/
/* iEnableDAC - Enable DAC Output (SpeakerOn)                                */
/*                                                                           */
/*****************************************************************************/
int     iEnableDAC( void )
	{
	int     iRetVal = _ES_DSP_ERROR;

	vDisplayDebugMessage( "iEnableDAC\n" );

	iRetVal = iWrDSP( 0xD1 );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iDisableDAC - Disable DAC Output (SpeakerOff)                             */
/*                                                                           */
/*****************************************************************************/
int     iDisableDAC( void )
	{
	int     iRetVal = _ES_DSP_ERROR;

	vDisplayDebugMessage( "iDisableDAC\n" );

	iRetVal = iWrDSP( 0xD3 );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iSetDSPTransferMode - Set dsp transfer modes                              */
/*                                                                           */
/*****************************************************************************/
int     iSetDSPTransferMode( BYTE _bDSPChannel, BYTE _bDSPModeFlags )
	{
	int     iRetVal = _ES_DSP_ERROR;

	vDisplayDebugMessage( "iSetDSPTransferMode = Channel:%02X Flags: %s/%s\n", _bDSPChannel, ( _bDSPModeFlags & _ES_AI ) ? "AutoInit" : "Non-AutoInit", ( _bDSPModeFlags & _ES_DEMAND8 ) ? "Demand" : "Single" );

	( _bDSPModeFlags & _ES_AI )			? iSetDSPAutoInitMode( _bDSPChannel )	: iSetDSPNonAutoInitMode( _bDSPChannel );
	( _bDSPModeFlags & _ES_DEMAND8 )	? iSetDSPDemandMode( _bDSPChannel )		: iSetDSPSingleMode( _bDSPChannel );
	( _bDSPModeFlags & _ES_16BIT )		? iSetDSP16BitTransfer( _bDSPChannel )	: iSetDSP8BitTransfer( _bDSPChannel );

	iRetVal = _ES_DSP_OK;

	return ( iRetVal );
	}

/*****************************************************************************/
/* iSetDSPTransferType - Set dsp transfer data type                          */
/*                                                                           */
/*****************************************************************************/
int     iSetDSPTransferType( BYTE _bDSPChannel, BYTE _bDSPTypeFlags )
	{
	int     iRetVal = _ES_DSP_ERROR;

	vDisplayDebugMessage( "iSetDSPTransferType = Channel:%02X\n", _bDSPChannel );

	( _bDSPTypeFlags & _ES_STEREO )	? iSetDSPStereo( _bDSPChannel )			: iSetDSPMono( _bDSPChannel );
	( _bDSPTypeFlags & _ES_SIGNED )	? iSetDSPSigned( _bDSPChannel )			: iSetDSPUnsigned( _bDSPChannel );
	( _bDSPTypeFlags & _ES_16BIT )	? iSetDSP16BitsPerSample( _bDSPChannel ): iSetDSP8BitsPerSample( _bDSPChannel );

	iRetVal = _ES_DSP_OK;
	return ( iRetVal );
	}

/*****************************************************************************/
/* iConfigureADC - Configure ADC for DSP input                               */
/*                                                                           */
/*****************************************************************************/
int     iConfigureDAC( BYTE _bDSPChannel, BYTE _bDSPTypeFlags )
	{
	int     iRetVal = _ES_DSP_ERROR;

    vDisplayDebugMessage( "iConfigureDAC = Channel:%02X Flags:%s/%s/%s\n", _bDSPChannel, 	\
    																		( _bDSPTypeFlags & _ES_16BIT ) ? "16bit" : "8bit", 
    																		( _bDSPTypeFlags & _ES_SIGNED ) ? "signed" : "unsigned", 
    																		( _bDSPTypeFlags & _ES_STEREO ) ? "stereo" : "mono" );
    

	switch( _bDSPChannel )
		{
		case _ES_CHANNEL0:
			if ( iWrXReg( 0xB6, ( BYTE ) ( ( _bDSPTypeFlags & _ES_16BIT ) ? 0x00 : 0x80 ) ) != _ES_DSP_ERROR )
				{
				if ( iWrXReg( 0xB7, ( BYTE ) ( ( _bDSPTypeFlags & _ES_SIGNED ) ? 0x71 : 0x51 ) ) != _ES_DSP_ERROR )
					{
					if ( iWrXReg( 0xB7, ( BYTE ) ( 0x90 | _bDSPTypeFlags ) ) != _ES_DSP_ERROR )
						{
						iRetVal = _ES_DSP_OK;
						}
					}
				}
			break;

		case _ES_CHANNEL1:
			iRetVal = _ES_DSP_OK;
			break;

		default:
			break;
		}
	return ( iRetVal );
	}

/*****************************************************************************/
/* iConfigureADC - Configure ADC for DSP input                               */
/*                                                                           */
/*****************************************************************************/
int     iConfigureADC( BYTE _bDSPChannel, BYTE _bDSPTypeFlags )
	{
	int     iRetVal = _ES_DSP_ERROR;

	vDisplayDebugMessage( "iConfigureADC = Channel:%02X Flags:%s/%s/%s\n", _bDSPChannel, 
																			( _bDSPTypeFlags & 0x04 ) ? "16bit" : "8bit", 
																			( _bDSPTypeFlags & 0x20 ) ? "signed" : "unsigned", 
																			( _bDSPTypeFlags & 0x40 ) ? "stereo" : "mono" );

	switch( _bDSPChannel )
		{
		case _ES_CHANNEL0:
			if ( iWrXReg( 0xB7, ( BYTE ) ( ( _bDSPTypeFlags & 0x20 ) ? 0x71 : 0x51 ) ) != _ES_DSP_ERROR )
				{
				if ( iWrXReg( 0xB7, ( BYTE ) ( 0x90 | _bDSPTypeFlags ) ) != _ES_DSP_ERROR )
					{
					iRetVal = _ES_DSP_OK;
					}
				}
			break;

		case _ES_CHANNEL1:
			iRetVal = _ES_DSP_OK;
			break;

		default:
			break;
		}
	return ( iRetVal );
	}

/*****************************************************************************/
/* iEnableFullDuplex - Enable simultaneous Playback (DMA2) & Record (DMA1)   */
/*                                                                           */
/*****************************************************************************/
int     iEnableFullDuplex( void )
	{
	int     iRetVal = _ES_DSP_OK;

	vDisplayDebugMessage( "iEnableFullDuplex:\t" );

	bSetMRegBit( _ES_DSP_MREG_ADAC_CTRL1, 0 );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iDisableFullDuplex - Disable Simultaneous Playback & Record               */
/*                                                                           */
/*****************************************************************************/
int     iDisableFullDuplex( void )
	{
	int     iRetVal = _ES_DSP_OK;

	vDisplayDebugMessage( "iDisableFullDuplex:\t" );

	bMskMRegBit( _ES_DSP_MREG_ADAC_CTRL1, 0 );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iEnableFullDuplex - Enable simultaneous Playback (DMA2) & Record (DMA1)   */
/*                                                                           */
/*****************************************************************************/
int     iEnableSecondaryDSPChannel( void )
	{
	int     iRetVal = _ES_DSP_OK;

	vDisplayDebugMessage( "iEnableSecondaryDSPChannel\t" );

	bSetMRegBit( _ES_DSP_MREG_ADAC_CTRL2, 5 );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iDisableFullDuplex - Disable Simultaneous Playback & Record               */
/*                                                                           */
/*****************************************************************************/
int     iDisableSecondaryDSPChannel( void )
	{
	int     iRetVal = _ES_DSP_OK;

	vDisplayDebugMessage( "iDisableSecondaryDSPChannel\t" );

	bMskMRegBit( _ES_DSP_MREG_ADAC_CTRL2, 5 );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iEnableFullDuplex - Enable simultaneous Playback (DMA2) & Record (DMA1)   */
/*                                                                           */
/*****************************************************************************/
int     iClearIRQEOutput( void )
	{
	int     iRetVal = _ES_DSP_OK;

	bMskMRegBit( _ES_DSP_MREG_ADAC_CTRL2, 7 );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iEnableFullDuplex - Enable simultaneous Playback (DMA2) & Record (DMA1)   */
/*                                                                           */
/*****************************************************************************/
int     iEnableIRQEOutput( void )
	{
	int     iRetVal = _ES_DSP_OK;

	vDisplayDebugMessage( "iEnableIRQEOutput\t" );

	bSetMRegBit( _ES_DSP_MREG_ADAC_CTRL2, 6 );

	return ( iRetVal );
	}

/*****************************************************************************/
/* iDisableFullDuplex - Disable Simultaneous Playback & Record               */
/*                                                                           */
/*****************************************************************************/
int     iDisableIRQEOutput( void )
	{
	int     iRetVal = _ES_DSP_OK;

	vDisplayDebugMessage( "iDisableIRQEOutput\t" );

	bMskMRegBit( _ES_DSP_MREG_ADAC_CTRL2, 6 );

	return ( iRetVal );
	}

