#include    <stdio.h>
#include	<stdlib.h>
#include	<bios.h>
#include	<dos.h>
#include	<conio.h>
#include    <math.h>

#include    "stdtype.h"
#include    "config.h"
#include    "delay.h"
#include    "fmsynth.h"


/* OPL2 I/O Port Definitions */
#define _FM_REG_STATUS  ( gwIOFMSynth + 0x00 )
#define _FM_REG_INDEX   ( gwIOFMSynth + 0x00 )
#define _FM_REG_DATA    ( gwIOFMSynth + 0x01 )


#define FM_MAX_CHANNEL  9

#define NOTECOUNT 6

BYTE gbFMRegs[0xFF];                                                                /* FM virtual register table */

BYTE gbModulatorCell[] ={  0x00, 0x01, 0x02, 0x08, 0x09, 0x0A, 0x10, 0x11, 0x12 };  /* Modulator Cells */

BYTE gbCarrierCell[] =  {  0x03, 0x04, 0x05, 0x0B, 0x0C, 0x0D, 0x13, 0x14, 0x15 };  /* Carrier Cells */


WORD p1[NOTECOUNT] = {0x0030, 0x0035, 0x0039, 0x003C, 0x0041, 0x0045};

                    /*   c       f       a       c       f       a   */
WORD p2[NOTECOUNT] = {0x02AE, 0x01CA, 0x0241, 0x02AE, 0x01CA, 0x0241};
BYTE p3[NOTECOUNT] = {0x0002, 0x0003, 0x0003, 0x0003, 0x0004, 0x0004};

typedef struct fm_op_struct
    {
    BYTE    ksl,
            fmult,
            fb,
            ar,
            sl,
            ss,
            dr,
            rr,
            tl,
            am,
            vib,
            ksr,
            con;
    } FM_OP;

typedef struct fm_ins_struct
    {
    char    mode;
    char    vnum;
    FM_OP   modulator;
    FM_OP   carrier;
    BYTE    ws_mod;
    BYTE    ws_car;
    } FM_INS;


FM_INS *fm_ins_ptr;


BYTE fm_ins[56] = { 0x00,
                    0x00,

                 /* Modulator Cell */
                    0x01, /* ksl  - (1)                         */
                    0x01, /* fmult- frequency multiplier (0)    */
                    0x00, /* fb   - feedback (0)                */
                    0x00, /* ar   - attack (15)                 */
                    0x00, /* sl   - sustain (5)                 */
                    0x01, /* ss   - envelope type (0)           */
                    0x00, /* dr   - decay (3)                   */
                    0x00, /* rr   - release (6)                 */
                    0x23, /* tl   - Attenuation (23)            */
                    0x01, /* am   - amplitude (on)              */
                    0x00, /* vib  - vibrato (off)               */
                    0x00, /* ksr  - ksr (off)                   */
                    0x01, /* con  - connection (type 1)         */

                 /* Carrier Cell */
                    0x02, /* ksl  - (2)                         */
                    0x01, /* fmult- frequency multiplier (1)    */
                    0x00, /* fb   - feedback (0)                */
                    0x0F, /* ar   - attack (15)                 */
                    0x00, /* sl   - sustain (5)                 */
                    0x01, /* ss   - envelope type (0)           */
                    0x0F, /* dr   - decay (2)                   */
                    0x04, /* rr   - release (4)                 */
                    0x00, /* tl   - Attenuation (0)             */
                    0x00, /* am   - amplitude (off)             */
                    0x00, /* vib  - vibrato (off)               */
                    0x00, /* ksr  - ksr (off)                   */
                    0x01, /* con  - connection (type 1)         */

                    0x01, 0x00

                    };


/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vWrFMRegIndex( BYTE _bRegIndex )
    {
    outp( _FM_REG_INDEX, _bRegIndex );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bRdFMRegIndex( void )
    {
    return ( (BYTE) ( inp( _FM_REG_INDEX ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vWrFMRegData( BYTE _bRegData )
    {
    outp( _FM_REG_DATA, _bRegData );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bRdFMRegData( void )
    {
    return ( (BYTE) ( inp( _FM_REG_DATA ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vWrFMReg( BYTE _bRegIndex, BYTE _bRegData )
    {
    BYTE bTimeout;

    vWrFMRegIndex( _bRegIndex );

//    vDelay( _1MS );

    vWrFMRegData( _bRegData );

//    vDelay( _1MS );

    gbFMRegs[_bRegIndex] = _bRegData;
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bRdFMReg( BYTE _bRegIndex )
    {
    return ( gbFMRegs[_bRegIndex] );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetFMRegBit( BYTE _bRegIndex, BYTE _bBit )
    {
    vWrFMReg( _bRegIndex, BSETBIT( bRdFMReg( _bRegIndex ), _bBit ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vMskFMRegBit( BYTE _bRegIndex, BYTE _bBit )
    {
    vWrFMReg( _bRegIndex, BMSKBIT( bRdFMReg( _bRegIndex ), _bBit ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bTstFMRegBit( BYTE _bRegIndex, BYTE _bBit )
    {
    return ( BTSTBIT( bRdFMReg( _bRegIndex ), _bBit ) );
    }


/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetCompositeModeEnable( BYTE _bState )
    {
    switch ( _bState )
        {
 case FALSE:
            vMskFMRegBit( 0x08, 7 );
            break;

        case TRUE:
            vSetFMRegBit( 0x08, 7 );
            break;
            }
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetCompositeModeEnable( void )
    {
    return ( bTstFMRegBit( 0x08, 7 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetKeyboardSplitEnable( BYTE _bState )
    {
    switch ( _bState )
        {
 case FALSE:
            vMskFMRegBit( 0x08, 6 );
            break;

        case TRUE:
            vSetFMRegBit( 0x08, 6 );
            break;
            }
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetKeyboardSplitEnable( void )
    {
    return ( bTstFMRegBit( 0x08, 6 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetAmplitudeEnable( BYTE _bCell, BYTE _bState )
    {
    switch ( _bState )
        {
 case FALSE:
            vMskFMRegBit( (BYTE) ( 0x20 + _bCell ), 7 );
            break;

        case TRUE:
            vSetFMRegBit( (BYTE) ( 0x20 + _bCell ), 7 );
            break;
            }
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetAmplitudeEnable( BYTE _bCell )
    {
    return ( bTstFMRegBit( (BYTE) ( 0x20 + _bCell ), 7 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetVibratoEnable( BYTE _bCell, BYTE _bState )
    {
    switch ( _bState )
        {
 case FALSE:
            vMskFMRegBit( (BYTE) ( 0x20 + _bCell ), 6 );
            break;

        case TRUE:
            vSetFMRegBit( (BYTE) ( 0x20 + _bCell ), 6 );
            break;
            }
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetVibratoEnable( BYTE _bCell )
    {
    return ( bTstFMRegBit( (BYTE) ( 0x20 + _bCell ), 6 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetEnvelopeType( BYTE _bCell, BYTE _bType )
    {
    switch ( _bType )
        {
 case FALSE:
            vMskFMRegBit( (BYTE) ( 0x20 + _bCell ), 5 );
            break;

        case TRUE:
            vSetFMRegBit( (BYTE) ( 0x20 + _bCell ), 5 );
            break;
            }
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetEnvelopeType( BYTE _bCell )
    {
    return ( bTstFMRegBit( (BYTE) ( 0x20 + _bCell ), 5 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetKSREnable( BYTE _bCell, BYTE _bState )
    {
    switch ( _bState )
        {
 case FALSE:
            vMskFMRegBit( (BYTE) ( 0x20 + _bCell ), 4 );
            break;

        case TRUE:
            vSetFMRegBit( (BYTE) ( 0x20 + _bCell ), 4 );
            break;
            }
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetKSREnable( BYTE _bCell )
    {
    return ( bTstFMRegBit( (BYTE) ( 0x20 + _bCell ), 4 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetMultiplier( BYTE _bCell, BYTE _bMultiplier )
    {
    vWrFMReg( (BYTE) ( 0x20 + _bCell ), (BYTE) ( ( bRdFMReg( (BYTE) ( 0x20 + _bCell ) ) & 0xF0 ) | ( _bMultiplier & 0x0F ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetMultiplier( BYTE _bCell )
    {
    return ( (BYTE) ( bRdFMReg( (BYTE) ( 0x20 + _bCell ) ) & 0x0F ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetTotalLevel( BYTE _bCell, BYTE _bLevel )
    {
    vWrFMReg( (BYTE) ( 0x40 + _bCell ), (BYTE) ( ( bRdFMReg( (BYTE) ( 0x40 + _bCell ) ) & 0xC0 ) | ( _bLevel & 0x3F ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetTotalLevel( BYTE _bCell )
    {
    return ( (BYTE) ( bRdFMReg( (BYTE) ( 0x40 + _bCell ) ) & 0x3F ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetKSL( BYTE _bCell, BYTE _bLevel )
    {
    vWrFMReg( (BYTE) ( 0x40 + _bCell ), (BYTE) ( ( bRdFMReg( (BYTE) ( 0x40 + _bCell ) ) & 0x7F ) | ( ( _bLevel & 0x03 ) << 6 ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetKSL( BYTE _bCell )
    {
    return ( (BYTE) ( ( bRdFMReg( (BYTE) ( 0x40 + _bCell ) ) & 0xC0 ) >> 6 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetAttackRate( BYTE _bCell, BYTE _bRate )
    {
    vWrFMReg( (BYTE) ( 0x60 + _bCell ), (BYTE) ( ( bRdFMReg( (BYTE) ( 0x60 + _bCell ) ) & 0x0F ) | ( _bRate << 4 ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetAttackRate( BYTE _bCell )
    {
    return ( (BYTE) ( bRdFMReg( (BYTE) ( 0x60 + _bCell ) ) >> 4 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetDecayRate( BYTE _bCell, BYTE _bRate )
    {
    vWrFMReg( (BYTE) ( 0x60 + _bCell ), (BYTE) ( ( bRdFMReg( (BYTE) ( 0x60 + _bCell ) ) & 0xF0 ) | ( _bRate & 0x0F ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetDecayRate( BYTE _bCell )
    {
    return ( (BYTE) ( bRdFMReg( (BYTE) ( 0x60 + _bCell ) ) & 0x0F ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetSustainLevel( BYTE _bCell, BYTE _bRate )
    {
    vWrFMReg( (BYTE) ( 0x80 + _bCell ), (BYTE) ( ( bRdFMReg( (BYTE) ( 0x80 + _bCell ) ) & 0xF0 ) | ( _bRate << 4 ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetSustainLevel( BYTE _bCell )
    {
    return ( (BYTE) ( bRdFMReg( (BYTE) ( 0x80 + _bCell ) ) >> 4 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetReleaseRate( BYTE _bCell, BYTE _bRate )
    {
    vWrFMReg( (BYTE) ( 0x80 + _bCell ), (BYTE) ( ( bRdFMReg( (BYTE) ( 0x80 + _bCell ) ) & 0xF0 ) | ( _bRate & 0x0F ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetReleaseRate( BYTE _bCell )
    {
    return ( (BYTE) ( bRdFMReg( (BYTE) ( 0x80 + _bCell ) ) & 0x0F ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetWaveSelect( BYTE _bCell, BYTE _bType )
    {
    vWrFMReg( (BYTE) ( 0xE0 + _bCell ), (BYTE) ( ( bRdFMReg( (BYTE) ( 0xE0 + _bCell ) ) & 0xFC ) | ( _bType & 0x03 ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetWaveSelect( BYTE _bCell )
    {
    return ( (BYTE) ( bRdFMReg( (BYTE) ( 0xE0 + _bCell ) ) & 0x03 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetConnectionType( BYTE _bChannel, BYTE _bType )
    {
    switch ( _bType )
        {
 case DISABLE:
            vMskFMRegBit( (BYTE) ( 0xC0 + _bChannel ), 0 );
            break;

        case ENABLE:
            vSetFMRegBit( (BYTE) ( 0xC0 + _bChannel ), 0 );
            break;
            }
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/

BYTE bGetConnectionType( BYTE _bChannel )
    {
    return ( bTstFMRegBit( (BYTE) ( 0xC0 + _bChannel ), 0 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetKeyEnable( BYTE _bChannel, BYTE _bState )
    {
    switch ( _bState )
        {
 case FALSE:
            vMskFMRegBit( (BYTE) ( 0xB0 + _bChannel ), 5 );
            break;

        case TRUE:
            vSetFMRegBit( (BYTE) ( 0xB0 + _bChannel ), 5 );
            break;
            }
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetKeyEnable( BYTE _bChannel )
    {
    return ( bTstFMRegBit( (BYTE) ( 0xB0 + _bChannel ), 5 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetFNumber( BYTE _bChannel, WORD _wFNumber )
    {
    vWrFMReg( (BYTE) ( 0xA0 + _bChannel ), LOBYTE( _wFNumber ) );
    vWrFMReg( (BYTE) ( 0xB0 + _bChannel ), (BYTE)( ( bRdFMReg( (BYTE)( 0xB0 + _bChannel ) ) & 0xFC ) | (BYTE)( HIBYTE( _wFNumber ) & 0x03 ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
WORD wGetFNumber( BYTE _bChannel )
    {
    return ( ( WORD ) ( ( bRdFMReg( (BYTE) ( 0xB0 + _bChannel ) ) & 0x03 ) << 8 ) | bRdFMReg( (BYTE) ( 0xA0 + _bChannel ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetBlockNumber( BYTE _bChannel, BYTE _bBlock )
    {
    vWrFMReg( (BYTE) ( 0xB0 + _bChannel ), (BYTE) ( ( bRdFMReg( (BYTE) ( 0xB0 + _bChannel ) ) & 0xE3 ) | ( ( _bBlock & 0x07 ) << 2 ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetBlockNumber( BYTE _bChannel )
    {
    return ( (BYTE) ( ( bRdFMReg( (BYTE) ( 0xB0 + _bChannel ) ) & 0x1C ) >> 2 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vSetFeedback( BYTE _bChannel, BYTE _bFeedback )
    {
    vWrFMReg( (BYTE) ( 0xC0 + _bChannel ), (BYTE) ( ( bRdFMReg( (BYTE) ( 0xC0 + _bChannel ) ) & 0xF1 ) | ( ( _bFeedback & 0x07 ) << 1 ) ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
BYTE bGetFeedback( BYTE _bChannel )
    {
    return ( (BYTE) ( ( bRdFMReg( (BYTE) ( 0xC0 + _bChannel ) ) & 0x0E ) >> 1 ) );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void vPlayESFMSynth( void )
    {
    int i;

    for ( i = 0; i < 4; i++ )
        {
        playsynth_sb(  );
        if ( kbhit(  ) )
            {
            getch(  );
            break;
            }
        }
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void    vPlaySineWave( void )
    {
    double  dTmp;

    WORD	wTable[] = { 363, 385, 408, 432, 458, 485, 514, 544, 577, 611, 647, 686 };

    BYTE    bBlock;

    WORD    wIndex;

    load_ins_data(  );                                                    /* Set AdLib Instrument */

    vSetCompositeModeEnable( ENABLE );

    vSetKeyEnable( 0x00, ENABLE );                                     /* Operator On */

    for( bBlock=0; bBlock < 8; bBlock++ )
        {
          for( wIndex=0; wIndex < 12; wIndex++ )
            {
            dTmp = ( 50000 * wTable[wIndex] ) * pow( (double)2, (double)( bBlock - 20 ) );

            vSetFNumber( 0x00, wTable[wIndex] );
            vSetBlockNumber( 0x00, bBlock );
            printf( "Block:%- 2u FNum:%- 3u = % -5.2lf\n", bBlock, wTable[wIndex], dTmp );

			vDelay( 0x1E848 );
            }
        }

    vSetKeyEnable( 0x00, DISABLE );

    vSetCompositeModeEnable( DISABLE );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void    vEnableSineWave( void )
    {
    load_ins_data(  );                                                    /* Set AdLib Instrument */

    vSetCompositeModeEnable( ENABLE );

    vSetFNumber( 0x00, 577 );
    vSetBlockNumber( 0x00, 3 );
    vSetKeyEnable( 0x00, ENABLE );                                     /* Operator On */
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void    vDisableSineWave( void )
    {
    vSetKeyEnable( 0x00, DISABLE );

    vSetCompositeModeEnable( DISABLE );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void playsynth_sb( void )
    {
    BYTE bChannel;

    vSetCompositeModeEnable( ENABLE );

    load_ins_data(  );                                                    /* Set AdLib Instrument */

    for ( bChannel = 0; bChannel < NOTECOUNT; ++bChannel )                     /* Turn on operators  */
        {
        vSetFNumber( 0x00, p2[bChannel] );
        vSetBlockNumber( 0x00, p3[bChannel] );
        vSetKeyEnable( 0x00, ENABLE );                                     /* Operator On */

        vDelay( 0x3D090 );

        vSetKeyEnable( 0x00, DISABLE );
        }

    vSetCompositeModeEnable( DISABLE );
    }

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/******************************************************************************/
void load_ins_data( void )
    {
    BYTE bChannel, bMCell, bCCell;

    fm_ins_ptr = ( FM_INS * ) & fm_ins;

    for ( bChannel = 0; bChannel < FM_MAX_CHANNEL; ++bChannel )
        {
        bMCell = gbModulatorCell[bChannel];
        bCCell = gbCarrierCell[bChannel];

        vSetAmplitudeEnable( bMCell, (BYTE) ( fm_ins_ptr->modulator.am ) );
        vSetVibratoEnable( bMCell, (BYTE) ( fm_ins_ptr->modulator.vib ) );
        vSetEnvelopeType( bMCell, (BYTE) ( fm_ins_ptr->modulator.ss ) );
        vSetKSREnable( bMCell, (BYTE) ( fm_ins_ptr->modulator.ksr ) );
        vSetMultiplier( bMCell, (BYTE) ( fm_ins_ptr->modulator.fmult ) );
        vSetKSL( bMCell, (BYTE) ( fm_ins_ptr->modulator.ksl ) );
        vSetTotalLevel( bMCell, (BYTE) ( fm_ins_ptr->modulator.tl ) );

        vSetAttackRate( bMCell, (BYTE) ( fm_ins_ptr->modulator.ar ) );
        vSetDecayRate( bMCell, (BYTE) ( fm_ins_ptr->modulator.dr ) );
        vSetSustainLevel( bMCell, (BYTE) ( fm_ins_ptr->modulator.sl ) );
        vSetReleaseRate( bMCell, (BYTE) ( fm_ins_ptr->modulator.rr ) );


        vSetAmplitudeEnable( bCCell, (BYTE) ( fm_ins_ptr->carrier.am ) );
        vSetVibratoEnable( bCCell, (BYTE) ( fm_ins_ptr->carrier.vib ) );
        vSetEnvelopeType( bCCell, (BYTE) ( fm_ins_ptr->carrier.ss ) );
        vSetKSREnable( bCCell, (BYTE) ( fm_ins_ptr->carrier.ksr ) );
        vSetMultiplier( bCCell, (BYTE) ( fm_ins_ptr->carrier.fmult ) );
        vSetKSL( bCCell, (BYTE) ( fm_ins_ptr->carrier.ksl ) );
        vSetTotalLevel( bCCell, (BYTE) ( fm_ins_ptr->carrier.tl ) );

        vSetAttackRate( bCCell, (BYTE) ( fm_ins_ptr->carrier.ar ) );
        vSetDecayRate( bCCell, (BYTE) ( fm_ins_ptr->carrier.dr ) );
        vSetSustainLevel( bCCell, (BYTE) ( fm_ins_ptr->carrier.sl ) );
        vSetReleaseRate( bCCell, (BYTE) ( fm_ins_ptr->carrier.rr ) );

        vSetFeedback( bChannel, (BYTE) ( fm_ins_ptr->modulator.fb ) );
        vSetConnectionType( bChannel, (BYTE) ( fm_ins_ptr->modulator.con ) );

        vSetWaveSelect( bMCell, (BYTE) ( fm_ins_ptr->ws_mod ) );
        vSetWaveSelect( bCCell, (BYTE) ( fm_ins_ptr->ws_car ) );
        }
    }
